# -*- coding: utf-8 -*-

"""
blueox.context
~~~~~~~~

This module provides the concept of 'Context' for collecting data that will
generate a log event.

:copyright: (c) 2012 by Rhett Garber
:license: ISC, see LICENSE for more details.

"""
import time
import os
import random
import struct
import threading
import functools

from . import utils
from . import network

# This will be redefined by our configuration to tell us where to record stuff
# to.
_recorder_function = None

class Context(object):
    __slots__ = ["name", "data", "id", "_writable", "start_time", "_sample_checks", "enabled"]
    def __init__(self, type_name, id=None, sample=None):

        if type_name.startswith('.'):
            parent_ctx = current_context()
            if parent_ctx is None:
                self.name = type_name[1:]
            else:
                parent_parts = sorted(enumerate(parent_ctx.name.split('.')), reverse=True)
                for ppart_ndx, ppart in parent_parts:
                    if ppart == type_name.split('.')[1]:
                        self.name = '.'.join(parent_ctx.name.split('.')[:ppart_ndx] + type_name.split('.')[1:])
                        break
                else:
                    self.name = parent_ctx.name + type_name
        else:
            parent_ctx = None
            self.name = type_name

        self.data = {}
        self.start_time = time.time()
        self._sample_checks = {}

        if id is not None:
            self.id = id
        elif parent_ctx:
            self.id = parent_ctx.id
        else:
            # Generate an id if one wasn't provided and we don't have any parents
            # We're going to encode the time as the front 4 bytes so we have some order to the ids
            # that could prove useful later on by making sorting a little easier.
            self.id = (struct.pack(">L", time.time()) + os.urandom(12)).encode('hex')

        if parent_ctx and not parent_ctx.enabled:
            self.enabled = False
        elif sample:
            sample_name, rate = sample
            if sample_name == type_name or sample_name == '.':
                self.enabled = bool(random.random() <= rate)
            elif parent_ctx and sample_name == '..':
                self.enabled = parent_ctx.sampled_for(type_name, rate)
            else:
                self.enabled = _get_context(sample_name).sampled_for(type_name, rate)
        else:
            self.enabled = True

        self._writable = False

    @property
    def writable(self):
        """Indicates the contest is open and can be written to"""
        return self._writable

    def sampled_for(self, name, rate):
        if name not in self._sample_checks:
            self._sample_checks[name] = bool(random.random() <= rate)
        return self._sample_checks[name]

    def set(self, key, *args, **kwargs):
        if not self.writable:
            raise ValueError()

        if args and kwargs:
            raise ValueError()

        if len(args) > 1:
            utils.set_deep(self.data, key, args)
        elif args:
            utils.set_deep(self.data, key, args[0])
        elif kwargs:
            existing_value = utils.get_deep(self.data, key, {})
            existing_value.update(kwargs)
            utils.set_deep(self.data, key, existing_value)

    def append(self, key, value):
        if not self.writable:
            raise ValueError()

        existing_value = utils.get_deep(self.data, key, [])
        existing_value.append(value)
        if len(existing_value) == 1:
            utils.set_deep(self.data, key, existing_value)

    def add(self, key, value):
        if not self.writable:
            raise ValueError()

        existing_value = utils.get_deep(self.data, key, 0)
        utils.set_deep(self.data, key, existing_value + value)

    def to_dict(self):
        return {'id': self.id,
                'type': self.name,
                'host': os.uname()[1],
                'pid': os.getpid(),
                'start': self.start_time,
                'end': time.time(),
                'body': self.data
               }

    def start(self):
        _add_context(self)
        self._writable = True

    def stop(self):
        self._writable = False
        _remove_context(self)

    def done(self):
        self.stop() # Just be sure
        if self.enabled and _recorder_function:
            _recorder_function(self)

        # Make sure we don't get any duplicate data
        # I would clear out all the data here, but that makes testing a little
        # more challenging.
        self.enabled = False

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, type, value, traceback):
        self.stop()
        self.done()

threadLocal = threading.local()

def init_contexts():
    if not getattr(threadLocal, 'init', None):
        threadLocal._contexts = []
        threadLocal._contexts_by_name = {}
        threadLocal.init = True

def _add_context(context):
    init_contexts()

    if context.name in threadLocal._contexts_by_name:
        return
    threadLocal._contexts_by_name[context.name] = context
    threadLocal._contexts.append(context)

def _get_context(name):
    return threadLocal._contexts_by_name.get(str(name))

def _remove_context(context):
    init_contexts()
    try:
        del threadLocal._contexts_by_name[context.name]
    except KeyError:
        pass

    try:
        threadLocal._contexts.remove(context)
    except ValueError:
        pass

def current_context():
    init_contexts()
    try:
        return threadLocal._contexts[-1]
    except IndexError:
        return None

def set(*args, **kwargs):
    context = current_context()
    if context:
        context.set(*args, **kwargs)

def append(*args, **kwargs):
    context = current_context()
    if context:
        context.append(*args, **kwargs)

def add(*args, **kwargs):
    context = current_context()
    if context:
        context.add(*args, **kwargs)

def context_wrap(type_name, sample=None):
    """Decorator for wrapping a function call with a context"""

    def wrapper(func):
        @functools.wraps(func)
        def inner(*args, **kwargs):
            with Context(type_name, sample=sample):
                return func(*args, **kwargs)

        return inner

    return wrapper



# -*- coding: utf-8 -*-

"""
blueox.tornado
~~~~~~~~

This module provides hooks for using blueox with the Tornado async web server.
Making blueox useful inside tornado is a challenge since you'll likely want a
blueox context per request, but multiple requests can be going on at once inside
tornado.

:copyright: (c) 2012 by Rhett Garber
:license: ISC, see LICENSE for more details.

"""
import functools
import logging
import traceback
import types

log = logging.getLogger(__name__)

import tornado.web
import tornado.gen
import tornado.simple_httpclient
import tornado.stack_context

import blueox

def install():
    """Install blueox hooks by poking into the tornado innards

    THIS MUST BE DONE BEFORE IMPORTING APPLICATION REQUEST HANDLERS
    We have to replace some decorators before they are used to create your class

    This is pretty hacky and may make you feel uncomfortable. It's only here so
    that blueox can be used with the minimal amount of extra boilerplate.  Your
    always free to be more explicit depending on your needs.

    Up to you if you want to hide your uglyiness in here or have it spread
    throughout your application.
    
    """
    tornado.gen.engine = gen_engine
    tornado.gen.Runner = BlueOxRunner


# Our hook into the request cycle is going to be provided by wrapping the
# _execute() method.  This creates our blueox context, starts it, and then stops
# it at the end. We'll leave it up to the finish() method to close us out.
def wrap_execute(type_name):
    def decorate(func):
        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            self.blueox = blueox.Context(type_name)
            self.blueox.start()
            try:
                return func(self, *args, **kwargs)
            finally:
                # We're done executing in this context for the time being. Either we've already
                # finished handling our blueox context, or we'll allow a later finish() call to 
                # mark it done.
                self.blueox.stop()

        return wrapper

    return decorate

def wrap_exception(func):
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        self.blueox.start()
        try:
            return func(self, *args, **kwargs)
        finally:
            self.blueox.stop()

    return wrapper

def wrap_finish(func):
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        try:
            return func(self, *args, **kwargs)
        finally:
            self.blueox.done()

    return wrapper

class SampleRequestHandler(tornado.web.RequestHandler):
    """Sample base request handler necessary for providing smart blueox contexts through a web request.

    We need to wrap two methods: _execute() and finish()

    To specify a name for your top level event, pass it the wrap_execute() decorator.

    The idea is that when _execute is called, we'll generate a blueox Context
    with the name specified. This should cover any methods like get() or
    post(). When _execute returns, we'll be in either one of two states. Either
    finish() has been called and we are all done, or finish() will be called
    later due to some async complexity. 

    Optionally, we also provide redefined methods that add critical data about
    the request to the active blueox context.
    """
    def prepare(self):
        blueox.set('headers', self.request.headers)
        blueox.set('method', self.request.method)
        blueox.set('uri', self.request.uri)

    def write_error(self, status_code, **kwargs):
        if 'exc_info' in kwargs:
            blueox.set('exception', ''.join(traceback.format_exception(*kwargs["exc_info"])))
    
        return super(SampleRequestHandler, self).write_error(status_code, **kwargs)

    def write(self, chunk):
        blueox.add('response_size', len(chunk))
        return super(SampleRequestHandler, self).write(chunk)

    def finish(self, *args, **kwargs):
        res = super(SampleRequestHandler, self).finish(*args, **kwargs)
        blueox.set('response_status_code', self._status_code)
        return res

    _execute = wrap_execute('request')(tornado.web.RequestHandler._execute)
    finish = wrap_finish(finish)
    _stack_context_handle_exception = wrap_exception(tornado.web.RequestHandler._stack_context_handle_exception)


# We need a custom version of this decorator so that we can pass in our blueox
# context to the Runner
def gen_engine(func):
    """Hacked up copy of tornado.gen.engine decorator
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        runner = None

        def handle_exception(typ, value, tb):
            # if the function throws an exception before its first "yield"
            # (or is not a generator at all), the Runner won't exist yet.
            # However, in that case we haven't reached anything asynchronous
            # yet, so we can just let the exception propagate.
            if runner is not None:
                return runner.handle_exception(typ, value, tb)
            return False

        with tornado.stack_context.ExceptionStackContext(handle_exception) as deactivate:
            gen = func(*args, **kwargs)
            if isinstance(gen, types.GeneratorType):
                blueox_ctx = getattr(args[0], 'blueox', None)
                runner = BlueOxRunner(gen, deactivate, blueox_ctx)
                runner.run()
                return
            assert gen is None, gen
            deactivate()
            # no yield, so we're done
    return wrapper


# Custom version of gen.Runner that starts and stops the blueox context
class BlueOxRunner(tornado.gen.Runner):
    def __init__(self, gen, deactivate_stack_context, blueox_context):
        self.blueox_ctx = blueox_context
        super(BlueOxRunner, self).__init__(gen, deactivate_stack_context)

    def run(self):
        try:
            if self.blueox_ctx:
                self.blueox_ctx.start()

            return super(BlueOxRunner, self).run()
        finally:
            if self.blueox_ctx:
                self.blueox_ctx.stop()

class AsyncHTTPClient(tornado.simple_httpclient.SimpleAsyncHTTPClient):
    def __init__(self, *args, **kwargs):
        self.blueox_name = '.httpclient'
        return super(AsyncHTTPClient, self).__init__(*args, **kwargs)

    def fetch(self, request, callback, **kwargs):
        ctx = blueox.Context(self.blueox_name)
        ctx.start()
        if isinstance(request, basestring):
            ctx.set('request.uri', request)
            ctx.set('request.method', kwargs.get('method', 'GET'))
        else:
            ctx.set('request.uri', request.url)
            ctx.set('request.method', request.method)
            ctx.set('request.size', len(request.body) if request.body else 0)

        ctx.stop()

        def wrap_callback(response):
            ctx.start()
            ctx.set('response.code', response.code)
            ctx.set('response.size', len(response.body) if response.body else 0)
            ctx.done()
            callback(response)

        return super(AsyncHTTPClient, self).fetch(request, wrap_callback, **kwargs)




import sys
import traceback
import logging

import blueox

from django.conf import settings

class Middleware:
    def __init__(self):
        host = getattr(settings, 'BLUEOX_HOST', '127.0.0.1')
        port = getattr(settings, 'BLUEOX_PORT', 3514)
        blueox.configure(host, port)

        handler = blueox.LogHandler()
        handler.setLevel(logging.INFO)
        logging.getLogger('').addHandler(handler)

    def process_request(self, request):
        request.blueox = blueox.Context(".".join((getattr(settings, 'BLUEOX_NAME', ''), 'request')))
        request.blueox.start()

        blueox.set('method', request.method)
        blueox.set('path', request.path)

        headers = {}
        for k,v in request.META.iteritems():
            if k.startswith('HTTP_') or k in ('CONTENT_LENGTH', 'CONTENT_TYPE'):
                headers[k] = v
        blueox.set('headers', headers)

        blueox.set('uri', request.build_absolute_uri())
        blueox.set('client_ip', request.META['REMOTE_ADDR'])

        for key in ('version', 'revision'):
            if hasattr(request, key):
                blueox.set(key, getattr(request, key))

        if request.user:
            blueox.set('user', request.user.id)

        return None

    def process_response(self, request, response):
        # process_request() is not guaranteed to be called
        if not hasattr(request, 'blueox'):
            return response

        # Other middleware may have blocked our response.
        if response is not None:
            blueox.set('response_status_code', response.status_code)

            if not response.streaming:
                blueox.set('response_size', len(response.content))

            headers = {}
            for k, v in response.items():
                headers[k] = v

            blueox.set('response_headers', headers)

        request.blueox.done()

        return response

    def process_exception(self, request, exception):
        blueox.set('exception', ''.join(traceback.format_exception(*sys.exc_info())))
        return None



# -*- coding: utf-8 -*-

"""
blueox
~~~~~~~~

:copyright: (c) 2012 by Rhett Garber
:license: ISC, see LICENSE for more details.

"""

__title__ = 'blueox'
__version__ = '0.5.10'
__author__ = 'Rhett Garber'
__author_email__ = 'rhettg@gmail.com'
__license__ = 'ISC'
__copyright__ = 'Copyright 2012 Rhett Garber'
__description__ = "A library for python-based application logging and data collection"
__url__ = "https://github.com/rhettg/BlueOx"

import logging

from . import utils
from . import network
from .context import Context, set, append, add, context_wrap
from . import context as _context_mod
from .errors import Error
from .logger import LogHandler
from .timer import timeit

log = logging.getLogger(__name__)


def configure(host, port, recorder=None):
    """Initialize blueox

    This instructs the blueox system where to send it's logging data. If blueox is not configured, log data will
    be silently dropped.

    Currently we support logging through the network (and the configured host and port) to a blueoxd instances, or
    to the specified recorder function
    """
    global _record_function
    if recorder:
        context._recorder_function = recorder
    elif host and port:
        network.init(host, port)
        context._recorder_function = network.send
    else:
        log.info("Empty blueox configuration")
        context._recorder_function = None

def shutdown():
    network.close()
